#include <iostream>
#include <fstream>
#include <time.h>  
#include <cmath>
#include <stdio.h>
#include <assert.h>
#include <cuda.h>
using namespace std;
using std::ifstream;

//#include <cutil.h>
//#include <cutil_math.h>
#include <limits.h>
#include <float.h>
#include <cfloat>
#include <vector>


//For reduction
//#include "cutil_inline.h"
#include <algorithm>
#include "cru_reduction_gpu.h"
//For reduction

#include "cru_global_defs.h"



#ifdef DO_UCLA_AP

 #include "cru_ucla.h"  

#endif
#include "HybridTaus.h"
#include "cru_diffusion_finitediff_GPU.h"

#include "cru_RyrUpdate_GPU.h"
/*
#ifdef TTdisrupt
 #include "cru_FluxUpdate_GPU_TTdisrupt.h"
 #include "cru_Lcc7stateUpdate_GPU_TTdisrupt.h"
 #include "cru_fun_maker_uniform_TTdisrupt.h"
#endif

#ifdef plcc  
 #include "cru_FluxUpdate_GPU_TTdisrupt.h" 
 #include "cru_Lcc7stateUpdate_GPU_TTdisrupt.h" 
 #include "cru_fun_maker_uniform_TTdisrupt_clusters.h"  
#endif
*/
#undef TTdisrupt
//#if !defined(TTdsirupt) && !defined(plcc)
 #include "cru_FluxUpdate_GPU.h"
 #include "cru_Lcc7stateUpdate_GPU.h"
 #include "cru_fun_maker_uniform.h"
//#endif

//GPU error checking macro
#define gpuErrchk(ans) { gpuAssert((ans), __FILE__, __LINE__); }
inline void gpuAssert(cudaError_t code, const char *file, int line, bool abort=true)
{
   if (code != cudaSuccess) 
   {
      fprintf(stderr,"GPUassert: %s %s %d\n", cudaGetErrorString(code), file, line);
      if (abort) exit(code);
   }
}



#ifdef BMOVIETIME
////////// FOR MOVIES //////////////////////////
FILE         *fptr_video;
FILE         *fptr_bounds;
unsigned char vm_frame_buffer[MAXBUFSIZE][REC_N_CELLS];
FLOAT         vm_bounds[MAXBUFSIZE][3];
static void write_frame(FLOAT t, FLOAT *x);
static void dump_data_to_files(void);

///////////////////////////////////////////////////////
#endif



int main (int argc, char** argv)
{
	if(argc == 2) {
		cudaSetDevice(atoi(argv[1]));
	} else {
		printf("Invalid arguments: device\n");
		return 1;
	}
 //cudaSetDevice(DevNum);
 const int threadsPerBlock=TperB;
 const int numBlocks = NxNyNz/threadsPerBlock + (NxNyNz%threadsPerBlock == 0?0:1);

 #ifdef BLOCK_SIZE_X
  dim3 dimBlock(BLOCK_SIZE_X,BLOCK_SIZE_Y); 
  dim3 dimGrid(Nx/dimBlock.x + (Nx%dimBlock.x == 0?0:1), Ny/dimBlock.y + (Ny%dimBlock.y == 0?0:1)  );
  //dim3 dimGrid((Nx + dimBlock.x - 1)/dimBlock.x, (Ny + dimBlock.y - 1)/dimBlock.y);
 #endif 

 //unsigned int timer = 0;
 int step=0;
 int usetemp=0;
 FLOAT Nai3value=Nai3;
 FLOAT t=0,v=vrest;
 #ifdef DO_FIXED_AP
  FLOAT tmod,tempnum1,tempnum2,tempnum3;
 #endif
 
 #ifdef DO_UCLA_AP
  Real x[NDIM];
  ucla_get_default_initial_state(x);
  FLOAT tmod;
  #ifndef USE_LCC_CELL
    FLOAT IlccTotal; 
  #endif
 #endif


#ifdef BMOVIETIME
 const FLOAT bmovietime1=BMOVIETIME,emovietime1=EMOVIETIME;
#endif


 size_t size = NxNyNz*sizeof(FLOAT);
 size_t sizeint=NxNyNz*sizeof(int);
 size_t sizeun=NxNyNz*sizeof(unsigned);

 FLOAT *cs; cs = (FLOAT *)malloc(size); memset(cs, 0, size);
 FLOAT *cstemp; cstemp=(FLOAT *)malloc(size);memset(cstemp, 0, size); 
 FLOAT *cm; cm=(FLOAT *)malloc(size);memset(cm, 0, size);
 FLOAT *cmtemp; cmtemp=(FLOAT *)malloc(size);memset(cmtemp, 0, size);
 FLOAT *cd; cd=(FLOAT *)malloc(size);memset(cd, 0, size);
 FLOAT *cdtemp; cdtemp=(FLOAT *)malloc(size);memset(cdtemp, 0, size);
 FLOAT *cj; cj=(FLOAT *)malloc(size);memset(cj, 0, size);
 FLOAT *cjtemp; cjtemp=(FLOAT *)malloc(size);memset(cjtemp, 0, size);
 
 
 FLOAT *cs_d; cudaMalloc((void **) &cs_d, size);
 FLOAT *cstemp_d; cudaMalloc((void **) &cstemp_d, size);
 FLOAT *cm_d; cudaMalloc((void **) &cm_d, size);
 FLOAT *cmtemp_d; cudaMalloc((void **) &cmtemp_d, size);
 #ifdef USE_TEX
 cudaBindTexture(NULL, tex_cm, cm_d, size);
 cudaBindTexture(NULL, tex_cmtemp, cmtemp_d, size);
 cudaBindTexture(NULL, tex_cs, cs_d, size);
 cudaBindTexture(NULL, tex_cstemp, cstemp_d, size);
 #endif
 
 FLOAT *cd_d; cudaMalloc((void **) &cd_d, size);
 FLOAT *cdtemp_d; cudaMalloc((void **) &cdtemp_d, size); 
 FLOAT *cj_d; cudaMalloc((void **) &cj_d, size);
 FLOAT *cjtemp_d; cudaMalloc((void **) &cjtemp_d, size);

 int *dfun; dfun=(int *)malloc(sizeint);memset(dfun, 0, sizeint);
 int *dfun_d; cudaMalloc((void **) &dfun_d, sizeint);


 FLOAT *Ilcc; Ilcc = (FLOAT *)malloc(size); memset(Ilcc, 0, size);
 FLOAT *Ilcc_d; cudaMalloc((void **) &Ilcc_d, size);
 FLOAT *Iryr; Iryr = (FLOAT *)malloc(size); memset(Iryr, 0, size);
 FLOAT *Iryr_d; cudaMalloc((void **) &Iryr_d, size);
 FLOAT *Jncx; Jncx=(FLOAT *)malloc(size);memset(Jncx, 0, size);
 FLOAT *Jncx_d; cudaMalloc((void **) &Jncx_d, size);

 FLOAT *Jup; Jup=(FLOAT *)malloc(size);memset(Jup, 0, size);
 FLOAT *Jleak; Jleak=(FLOAT *)malloc(size);memset(Jleak, 0, size);
 FLOAT *FluxS; FluxS=(FLOAT *)malloc(size);memset(FluxS, 0, size);
 FLOAT *FluxM; FluxM=(FLOAT *)malloc(size);memset(FluxM, 0, size);
 FLOAT *FluxJ; FluxJ=(FLOAT *)malloc(size);memset(FluxJ, 0, size);
 FLOAT *FluxD; FluxD=(FLOAT *)malloc(size);memset(FluxD, 0, size);


 FLOAT *Pro; Pro=(FLOAT *)malloc(size);memset(Pro, 0, size);
 FLOAT *Pro_d; cudaMalloc((void **) &Pro_d, size);
 FLOAT *Pri; Pri=(FLOAT *)malloc(size);memset(Pri, 0, size);
 FLOAT *Pri_d; cudaMalloc((void **) &Pri_d, size);
 FLOAT *Prc; Prc=(FLOAT *)malloc(size);memset(Prc, 0, size);
 FLOAT *Prc_d; cudaMalloc((void **) &Prc_d, size);
 FLOAT *Protemp; Protemp=(FLOAT *)malloc(size);memset(Protemp, 0, size);
 FLOAT *Protemp_d; cudaMalloc((void **) &Protemp_d, size);
 FLOAT *Pritemp; Pritemp=(FLOAT *)malloc(size);memset(Pritemp, 0, size);
 FLOAT *Pritemp_d; cudaMalloc((void **) &Pritemp_d, size);
 FLOAT *Prctemp; Prctemp=(FLOAT *)malloc(size);memset(Prctemp, 0, size);
 FLOAT *Prctemp_d; cudaMalloc((void **) &Prctemp_d, size);

 FLOAT *SUMdumb; SUMdumb=(FLOAT *)malloc(size);memset(SUMdumb, 0, size);
 FLOAT *SUMdumb_d; cudaMalloc((void **) &SUMdumb_d, size);

 int *nr; nr=(int *)malloc(sizeint);memset(nr, 0, sizeint);
 int *nr_d; cudaMalloc((void **) &nr_d, sizeint);

 int *Nr; Nr=(int *)malloc(sizeint);memset(Nr, 0, sizeint);
 int *Nri; Nri=(int *)malloc(sizeint);memset(Nri, 0, sizeint);
 int *Nrc; Nrc=(int *)malloc(sizeint);memset(Nrc, 0, sizeint);
 int *Nrr; Nrr=(int *)malloc(sizeint);memset(Nrr, 0, sizeint);
 int *Nr_d; cudaMalloc((void **) &Nr_d, sizeint);
 int *Nrc_d; cudaMalloc((void **) &Nrc_d, sizeint);
 int *Nrr_d; cudaMalloc((void **) &Nrr_d, sizeint);
 int *Nri_d; cudaMalloc((void **) &Nri_d, sizeint);

 int *sumint_d; cudaMalloc((void **) &sumint_d, sizeint);
 FLOAT *sumfloat_d; cudaMalloc((void **) &sumfloat_d, size);

 FLOAT *rfire; rfire=(FLOAT *)malloc(size);memset(rfire, 0, size);
 FLOAT *rfire_d; cudaMalloc((void **) &rfire_d, size);
 //double Nravg,Nrcavg;


 int *nl; nl=(int *)malloc(sizeint);memset(nl, 0, sizeint);

 int *Nl; Nl=(int *)malloc(sizeint);memset(Nl, 0, sizeint);
 int *Nli; Nli=(int *)malloc(sizeint);memset(Nli, 0, sizeint);
 int *Nlc; Nlc=(int *)malloc(sizeint);memset(Nlc, 0, sizeint);
 int *Nli2; Nli2=(int *)malloc(sizeint);memset(Nli2, 0, sizeint);
 int *Nlc2; Nlc2=(int *)malloc(sizeint);memset(Nlc2, 0, sizeint);
 int *Nlb1; Nlb1=(int *)malloc(sizeint);memset(Nlb1, 0, sizeint);
 int *Nlb2; Nlb2=(int *)malloc(sizeint);memset(Nlb2, 0, sizeint);
 int *Nl_d; cudaMalloc((void **) &Nl_d, sizeint);
 int *Nli_d; cudaMalloc((void **) &Nli_d, sizeint);
 int *Nli2_d; cudaMalloc((void **) &Nli2_d, sizeint);
 int *Nlc_d; cudaMalloc((void **) &Nlc_d, sizeint);
 int *Nlc2_d; cudaMalloc((void **) &Nlc2_d, sizeint);
 int *Nlb1_d; cudaMalloc((void **) &Nlb1_d, sizeint);
 int *Nlb2_d; cudaMalloc((void **) &Nlb2_d, sizeint);

 FLOAT *lfire; lfire=(FLOAT *)malloc(size);memset(lfire, 0, size);
 FLOAT *lfire_d; cudaMalloc((void **) &lfire_d, size);
 //double Nlavg,Nlcavg; 

 //double cmavg,csavg,cdavg,cjavg,Nlavg,Nravg,Nrcavg,Jncxavg,Ilccavg,Iryravg;



 int *spark; spark=(int *)malloc(sizeint);memset(spark, 0, sizeint);
 int *fspark; fspark=(int *)malloc(sizeint);memset(fspark, 0, sizeint);
 FLOAT *sparkT; sparkT=(FLOAT *)malloc(size);memset(sparkT, 0, size);



 unsigned *seed_h1; seed_h1=( unsigned *)malloc(sizeun);memset(seed_h1, 0, sizeun);
 unsigned *seed_d1; cudaMalloc((void **) &seed_d1, sizeun);
 unsigned *seed_h2; seed_h2=(unsigned *)malloc(sizeun);memset(seed_h2, 0, sizeun);
 unsigned *seed_d2; cudaMalloc((void **) &seed_d2, sizeun);
 unsigned *seed_h3; seed_h3=(unsigned *)malloc(sizeun);memset(seed_h3, 0, sizeun);
 unsigned *seed_d3; cudaMalloc((void **) &seed_d3, sizeun);
 unsigned *seed_h4; seed_h4=(unsigned *)malloc(sizeun);memset(seed_h4, 0, sizeun);
 unsigned *seed_d4; cudaMalloc((void **) &seed_d4, sizeun);
 for (int k=0; k<Nz; k++){
  for (int i=0; i<Ny; i++){
   for (int j=0; j<Nx; j++){
    int idx=Nx*Ny*k +Nx*i + j;
     seed_h1[idx]=(int)time(0)*(idx+1);
     seed_h2[idx]=(int)time(0)*(idx+1);
     seed_h3[idx]=(int)time(0)*(idx+1);
     seed_h4[idx]=(int)time(0)*(idx+1);
   };
  };
 };
 cudaMemcpy(seed_d1, seed_h1, sizeun, cudaMemcpyHostToDevice); 
 cudaMemcpy(seed_d2, seed_h2, sizeun, cudaMemcpyHostToDevice); 
 cudaMemcpy(seed_d3, seed_h3, sizeun, cudaMemcpyHostToDevice); 
 cudaMemcpy(seed_d4, seed_h4, sizeun, cudaMemcpyHostToDevice); 




 ofstream avgfile,sparkfile,singlesparkfile;

#ifdef AP_CLAMP_FILE
 ofstream apfile;
 apfile.open(AP_CLAMP_FILE);
#endif

#ifdef AP_CLAMP
ifstream infile;
infile.open(AP_CLAMP);
if(!infile) { // file couldn't be opened
 cerr << "Error: file could not be opened" << endl;
 exit(1);
};
FLOAT tdum,vdum;
vector<FLOAT> v_apclamp;
while(! infile.eof()){
 infile >> tdum >> vdum;
  v_apclamp.push_back(vdum);
};
v=v_apclamp[0];
int length_v_apclamp=v_apclamp.size();//(PCL/AP_OUTTIME)-1;
float vaprest=v_apclamp[0];
//cout<<"length_v_apclamp" << length_v_apclamp<<"vaprest" << vaprest<<endl;
float aptmod;
#endif

 avgfile.open(AVG_FILE);
 sparkfile.open(SPARK_FILE);
 singlesparkfile.open(SINGLE_SPARK_FILE);
 

#ifdef BMOVIETIME
////////// FOR MOVIES //////////////////////////
  fptr_video  = fopen(MOVIE_FILE, "wb");
  assert(fptr_video != NULL);
  fptr_bounds = fopen(BOUNDS_FILE, "w");
  assert(fptr_bounds != NULL);
///////////////////////////////////////////////////
#endif

 step=0;
 t=0;
 


//initialize//////////////////////////////////////////
int crucount=0;


fun_maker_uniform(cm, cs, cd, cj, nr, nl, Nr, Nrc, Nri, Nrr, Nl, Nlc, Nlc2, Nli, Nli2, Nlb1, Nlb2, rfire, lfire, dfun, crucount, Pro, Pri, Prc, Ilcc, Iryr);

cout<<"crucount" << crucount<<endl;


int lcccount=0;
for (int k=0; k<Nz; k++){
 
 for (int i=0; i<Ny; i++){
   
  for (int j=0; j<Nx; j++){
   int idx=Nx*Ny*k +Nx*i + j;
   

   lcccount+=nl[idx];

   

  };
 };
};

cout<<"lcccount" << lcccount<<endl;




cudaMemcpy(cm_d, cm, size, cudaMemcpyHostToDevice);
cudaMemcpy(cmtemp_d, cm, size, cudaMemcpyHostToDevice);
cudaMemcpy(cs_d, cs, size, cudaMemcpyHostToDevice);
cudaMemcpy(cstemp_d, cs, size, cudaMemcpyHostToDevice);
cudaMemcpy(cd_d, cd, size, cudaMemcpyHostToDevice);
cudaMemcpy(cdtemp_d, cd, size, cudaMemcpyHostToDevice);
cudaMemcpy(cj_d, cj, size, cudaMemcpyHostToDevice);
cudaMemcpy(cjtemp_d, cj, size, cudaMemcpyHostToDevice);

cudaMemcpy(dfun_d, dfun, sizeint, cudaMemcpyHostToDevice);

cudaMemcpy(sumint_d, dfun, sizeint, cudaMemcpyHostToDevice);

cudaMemcpy(Nl_d, Nl, sizeint, cudaMemcpyHostToDevice);
cudaMemcpy(Nlc_d, Nlc, sizeint, cudaMemcpyHostToDevice);
cudaMemcpy(Nli_d, Nli, sizeint, cudaMemcpyHostToDevice);
cudaMemcpy(Nlb1_d, Nlb1, sizeint, cudaMemcpyHostToDevice);
cudaMemcpy(Nlb2_d, Nlb2, sizeint, cudaMemcpyHostToDevice);
cudaMemcpy(Nlc2_d, Nlc2, sizeint, cudaMemcpyHostToDevice);
cudaMemcpy(Nli2_d, Nli2, sizeint, cudaMemcpyHostToDevice);
cudaMemcpy(lfire_d, lfire, size, cudaMemcpyHostToDevice);

cudaMemcpy(Nr_d, Nr, sizeint, cudaMemcpyHostToDevice);
cudaMemcpy(Nrr_d, Nrr, sizeint, cudaMemcpyHostToDevice);
cudaMemcpy(Nrc_d, Nrc, sizeint, cudaMemcpyHostToDevice);
cudaMemcpy(Nri_d, Nri, sizeint, cudaMemcpyHostToDevice);
cudaMemcpy(rfire_d, rfire, size, cudaMemcpyHostToDevice);
cudaMemcpy(nr_d, nr, sizeint, cudaMemcpyHostToDevice);


cudaMemcpy(Pro_d, Pro, size, cudaMemcpyHostToDevice);
cudaMemcpy(Pri_d, Pri, size, cudaMemcpyHostToDevice);
cudaMemcpy(Prc_d, Prc, size, cudaMemcpyHostToDevice);
cudaMemcpy(Protemp_d, Pro, size, cudaMemcpyHostToDevice);
cudaMemcpy(Pritemp_d, Pri, size, cudaMemcpyHostToDevice);
cudaMemcpy(Prctemp_d, Prc, size, cudaMemcpyHostToDevice);

cudaMemcpy(SUMdumb_d, Pro, size, cudaMemcpyHostToDevice);


cudaMemcpy(Jncx_d, Jncx, size, cudaMemcpyHostToDevice);
cudaMemcpy(Ilcc_d, Ilcc, size, cudaMemcpyHostToDevice);
cudaMemcpy(Iryr_d, Iryr, size, cudaMemcpyHostToDevice);




//unsigned int timer = 0;
/*cutCreateTimer( &timer);
cutStartTimer( timer);*/
cout<<"start"<<" "<<tottime<<" "<< avgfile << " "<< "DevNum"<< " " << DevNum <<endl;
cout<<"dt"<<" "<<dt<<endl;
cout<<"dtflux"<<" "<<dtflux<<endl;
cout<<"dtdiff"<<" "<<dtdiff<<endl;



//////// end initialize //////////////
FILE* outFile = fopen("test.txt","w");

while (step<step_tot){


////////output/////////////////////////////////////

#ifdef AP_CLAMP_FILE
if (t>=AP_REC_START && t<=AP_REC_STOP){
 if (step % AP_OUTSTEP == 0){
  apfile << t << " "<< v<<endl;
 };
};
#endif

///////////start fast output///////
#ifdef DO_OUTPUT_FAST
if (step % outstep == 0){

 double cmavg;
 double csavg;
 if (usetemp == 1){
 //cudaMemcpy(cm, cmtemp_d, size, cudaMemcpyDeviceToHost);
 cmavg = computeReduction(NxNyNz, cmtemp_d, SUMdumb_d);
 csavg = computeReduction(NxNyNz, cstemp_d, SUMdumb_d);
 }
 else{
 //cudaMemcpy(cm, cm_d, size, cudaMemcpyDeviceToHost);
 cmavg = computeReduction(NxNyNz, cm_d, SUMdumb_d);
 csavg = computeReduction(NxNyNz, cs_d, SUMdumb_d);
 };

 double cdavg = computeReduction(NxNyNz, cd_d, SUMdumb_d); 
 double cjavg = computeReduction(NxNyNz, cj_d, SUMdumb_d); 
 double Nlavg = computeReductionINT(NxNyNz, Nl_d, sumint_d);
 double Nlc2avg=computeReductionINT(NxNyNz, Nlc2_d, sumint_d);
 double Nravg = computeReductionINT(NxNyNz, Nr_d, sumint_d);
 double Nrcavg = computeReductionINT(NxNyNz, Nrc_d, sumint_d);
 double Jncxavg = computeReduction(NxNyNz, Jncx_d, SUMdumb_d);
 double Ilccavg = computeReduction(NxNyNz, Ilcc_d, SUMdumb_d);
 double Iryravg = computeReduction(NxNyNz, Iryr_d, SUMdumb_d);


 cmavg=cmavg/(NxNyNz);
 csavg=csavg/(NxNyNz);
 Jncxavg=Jncxavg/(NxNyNz);
 cdavg=cdavg/(crucount);
 cjavg=cjavg/(crucount);
 Nravg=Nravg/(crucount);
 Nrcavg=Nrcavg/(crucount);
 Nlavg=Nlavg/(crucount);
 Nlc2avg=Nlc2avg/(crucount);
 Ilccavg=Ilccavg/(crucount);
 Iryravg=Iryravg/(crucount);

 #ifdef DO_UCLA_AP

avgfile << t <<" "<<cmavg<<" "<<csavg<<" "<<cdavg<<" "<<cjavg<<" "<<Nravg<<" "<<Nrcavg<<" " <<Nlavg<<"  "<< Nlc2avg << " " <<Ilccavg <<" "<<x[28]<<" "<<Iryravg<<" "<<" "<<x[27]<<" "<<v<<endl;
 
#else

  //avgfile << t <<" "<<cmavg<<" "<<csavg<<" "<<cdavg<<" "<<cjavg<<" "<<Nravg<<" "<<Nrcavg<<" " <<Nlavg<<" "<<Nlcavg<<endl;
 avgfile << t <<" "<<cmavg<<" "<<csavg<<" "<<cdavg<<" "<<cjavg<<" "<<Nravg<<" "<<Nrcavg<<" " <<Nlavg<<" "<< Nlc2avg << " "<<Ilccavg<<" "<<Iryravg<<" "<<Jncxavg<<" "<<v<<endl;

#endif 

#ifdef BMOVIETIME
 if(t>=bmovietime1 && t<=emovietime1 && step % moviestep == 0 ){

  
  ////////// FOR MOVIES //////////////////////////
 #ifdef SR_MOVIE
   write_frame(t, cs);
 #else
   write_frame(t, cm);
 #endif
   dump_data_to_files();
  ////////////////////////////////////////////////

 };
#endif 



 #ifdef COMPUTE_SPARK_T
  cudaMemcpy(Nr, Nr_d, sizeint, cudaMemcpyDeviceToHost); 
  
  for (int k=0; k<Nz; k++){
   for (int i=0; i<Ny; i++){
    for (int j=0; j<Nx; j++){
     int idx=Nx*Ny*k +Nx*i + j;
 
  if(  Nr[idx]>=SPARK_U_T && spark[idx]==0 && fspark[idx]==1 && t> COMPUTE_SPARK_T ){ sparkfile<<t<<" "<< t-sparkT[idx]<<" "<<i<<" "<<j<<" "<<k<<endl; sparkT[idx]=t; spark[idx]=1;}

  else if ( Nr[idx]>=SPARK_U_T && spark[idx]==0 && fspark[idx]==0 && t> COMPUTE_SPARK_T ) { sparkfile<<t<<" "<< 0 <<" "<<i<<" "<<j<<" "<<k<<endl; sparkT[idx]=t; spark[idx]=1; fspark[idx]=1; } 

  else if ( Nr[idx]<=SPARK_L_T ){ spark[idx]=0;};
     };
    };
   };
  
 #endif

 #ifdef Spark_Plot
  #ifndef COMPUTE_SPARK_T
   cudaMemcpy(Nr, Nr_d, sizeint, cudaMemcpyDeviceToHost); 
  #endif
  
  singlesparkfile << t << " "<< Nr[Nx*Ny*SparkZ +Nx*SparkY + SparkX] << " "<< Nr[Nx*Ny*(SparkZ2) +Nx*(SparkY2) + SparkX2] <<endl;

 #endif

};
#endif
////////////////end fast output


#ifdef DO_OUTPUT
if (step % outstep == 0){

if (usetemp == 1){
 cudaMemcpy(cm, cmtemp_d, size, cudaMemcpyDeviceToHost);
 cudaMemcpy(cs, cstemp_d, size, cudaMemcpyDeviceToHost);
}
else{
 cudaMemcpy(cm, cm_d, size, cudaMemcpyDeviceToHost);
 cudaMemcpy(cs, cs_d, size, cudaMemcpyDeviceToHost);
};


 cudaMemcpy(cd, cd_d, size, cudaMemcpyDeviceToHost);
 cudaMemcpy(cj, cj_d, size, cudaMemcpyDeviceToHost);

 cudaMemcpy(Nr, Nr_d, sizeint, cudaMemcpyDeviceToHost);
 cudaMemcpy(Nrc, Nrc_d, sizeint, cudaMemcpyDeviceToHost);
 cudaMemcpy(Pro, Pro_d, size, cudaMemcpyDeviceToHost);
 cudaMemcpy(Prc, Prc_d, size, cudaMemcpyDeviceToHost);

 cudaMemcpy(Nl, Nl_d, sizeint, cudaMemcpyDeviceToHost);
 cudaMemcpy(Nlc, Nlc_d, sizeint, cudaMemcpyDeviceToHost);
 cudaMemcpy(Nlc2, Nlc2_d, sizeint, cudaMemcpyDeviceToHost);
 cudaMemcpy(Nli, Nli_d, sizeint, cudaMemcpyDeviceToHost);
 cudaMemcpy(Nli2, Nli2_d, sizeint, cudaMemcpyDeviceToHost);
 cudaMemcpy(Nlb1, Nlb1_d, sizeint, cudaMemcpyDeviceToHost);
 cudaMemcpy(Nlb2, Nlb2_d, sizeint, cudaMemcpyDeviceToHost);

 cudaMemcpy(Ilcc, Ilcc_d, size, cudaMemcpyDeviceToHost);
 cudaMemcpy(Iryr, Iryr_d, size, cudaMemcpyDeviceToHost);
 cudaMemcpy(Jncx, Jncx_d, size, cudaMemcpyDeviceToHost);

  /*
   cudaError_t err2 = cudaGetLastError();
   if( err2 != cudaSuccess ) {
    cout<<step<<endl;
    printf("Error: %s\n", cudaGetErrorString(err2) );
    assert(err2== cudaSuccess);
   };
  */

 double cmavg=0.0;
 double csavg=0.0;
 double cdavg=0.0;
 double cjavg=0.0;

 double Nravg=0.0;
 double Nrcavg=0.0;
 double Proavg=0.0;
 double Prcavg=0.0;

 double Nlavg=0.0;
 double Nlcavg=0.0;
 double Nlc2avg=0.0;
 double Nliavg=0.0;
 double Nli2avg=0.0;
 double Nlb1avg=0.0;
 double Nlb2avg=0.0; 

 double Ilccavg=0.0;
 double Iryravg=0.0;
 double Jncxavg=0.0;


for (int k=0; k<Nz; k++){
 
 for (int i=0; i<Ny; i++){
   
  for (int j=0; j<Nx; j++){
   int idx=Nx*Ny*k +Nx*i + j;
   
  // if (t==0 && Nrc[idx]!=0) {cout<<Nrc[idx]<<endl;}; 

   cmavg+=cm[idx];
   csavg+=cs[idx]; 
   cdavg+=cd[idx];
   cjavg+=cj[idx];
   
   Nravg+=Nr[idx];
   Nrcavg+=Nrc[idx];
   Nlavg+=Nl[idx];
   Nlcavg+=Nlc[idx];
   Nlc2avg+=Nlc2[idx];
   Nliavg+=Nli[idx];
   Nli2avg+=Nli2[idx];
   Nlb1avg+=Nlb1[idx];
   Nlb2avg+=Nlb2[idx];

   Ilccavg+=Ilcc[idx];
   Iryravg+=Iryr[idx];
   Jncxavg+=Jncx[idx];
   Proavg+=Pro[idx];
   Prcavg+=Prc[idx];
  
  #ifdef COMPUTE_SPARK_T
  
  if(  Nr[idx]>=SPARK_U_T && spark[idx]==0 && fspark[idx]==1 && t> COMPUTE_SPARK_T ){ sparkfile<<t<<" "<< t-sparkT[idx]<<" "<<i<<" "<<j<<" "<<k<<endl; sparkT[idx]=t; spark[idx]=1;}

  else if ( Nr[idx]>=SPARK_U_T && spark[idx]==0 && fspark[idx]==0 && t> COMPUTE_SPARK_T ) { sparkfile<<t<<" "<< 0 <<" "<<i<<" "<<j<<" "<<k<<endl; sparkT[idx]=t; spark[idx]=1; fspark[idx]=1; } 

  else if ( Nr[idx]<=SPARK_L_T ){ spark[idx]=0;};

  #endif

  };
 };
 };
 

 cmavg=cmavg/(NxNyNz);
 csavg=csavg/(NxNyNz);
 Jncxavg=Jncxavg/(NxNyNz);
 cdavg=cdavg/(crucount);
 cjavg=cjavg/(crucount);
 Nravg=Nravg/(crucount);
 Nrcavg=Nrcavg/(crucount);
 Nlavg=Nlavg/(crucount);
 Nlcavg=Nlcavg/(crucount);
 Nlc2avg=Nlc2avg/(crucount);
 Nliavg=Nliavg/(crucount);
 Nli2avg=Nli2avg/(crucount);
 Nlb1avg=Nlb1avg/(crucount);
 Nlb2avg=Nlb2avg/(crucount);
 Ilccavg=Ilccavg/(crucount);
 Iryravg=Iryravg/(crucount);
 Proavg=Proavg/(crucount);
 Prcavg=Prcavg/(crucount);

#ifdef USE_RYR_ODE 
Nravg=Proavg;
Nrcavg=Prcavg;
#endif

#ifdef DO_UCLA_AP

avgfile << t <<" "<<cmavg<<" "<<csavg<<" "<<cdavg<<" "<<cjavg<<" "<<Nravg<<" "<<Nrcavg<<" " <<Nlavg<<" "<<Nlcavg<<" "<<Nlc2avg <<  " "<< Nliavg <<  " "<<Nli2avg << " "<<Nlb1avg <<" "<<" "<<Ilccavg <<" "<<x[28]<<" "<<Iryravg<<" "<<" "<<x[27]<<" "<<v<<" "<<x[19]<<endl;
 
#else

  //avgfile << t <<" "<<cmavg<<" "<<csavg<<" "<<cdavg<<" "<<cjavg<<" "<<Nravg<<" "<<Nrcavg<<" " <<Nlavg<<" "<<Nlcavg<<endl;
 avgfile << t <<" "<<cmavg<<" "<<csavg<<" "<<cdavg<<" "<<cjavg<<" "<<Nravg<<" "<<Nrcavg<<" " <<Nlavg<<" "<<Nlcavg<<" "<<Nlc2avg <<  " "<< Nliavg <<  " "<<Nli2avg << " "<<Nlb1avg <<" "<<Nlb2avg <<" "<<Ilccavg<<" "<<Iryravg<<" "<<Jncxavg<<" "<<v<<endl;

#endif 

 #ifdef Spark_Plot
 
 #ifdef OUTPUT_ALL

 cudaMemcpy(Nrr, Nrr_d, sizeint, cudaMemcpyDeviceToHost);
 cudaMemcpy(Nri, Nri_d, sizeint, cudaMemcpyDeviceToHost);

 singlesparkfile << t <<" "<<SparkX<<" "<< cm[Nx*Ny*SparkZ +Nx*SparkY + SparkX]<<" "<< cs[Nx*Ny*SparkZ +Nx*SparkY + SparkX]<< " "<< cd[Nx*Ny*SparkZ +Nx*SparkY + SparkX]<<" "<< cj[Nx*Ny*SparkZ +Nx*SparkY + SparkX]<< " "<< Nr[Nx*Ny*SparkZ +Nx*SparkY + SparkX]<< " "<< Nrc[Nx*Ny*SparkZ +Nx*SparkY + SparkX]<< " " << Nri[Nx*Ny*SparkZ +Nx*SparkY + SparkX] << " " << Nrr[Nx*Ny*SparkZ +Nx*SparkY + SparkX] << " " << Nl[Nx*Ny*SparkZ +Nx*SparkY + SparkX]<< " " << Nlc[Nx*Ny*SparkZ +Nx*SparkY + SparkX]<< " " << Nlc2[Nx*Ny*SparkZ +Nx*SparkY + SparkX]<< " " << Nli[Nx*Ny*SparkZ +Nx*SparkY + SparkX] << " " << Nli2[Nx*Ny*SparkZ +Nx*SparkY + SparkX]<< " " << Nlb1[Nx*Ny*SparkZ +Nx*SparkY + SparkX]<< " " << Nlb2[Nx*Ny*SparkZ +Nx*SparkY + SparkX]<<" "<<SparkX2<<" "<< cm[Nx*Ny*SparkZ2 +Nx*SparkY2 + SparkX2]<<" "<< cs[Nx*Ny*SparkZ2 +Nx*SparkY2 + SparkX2]<< " "<< cd[Nx*Ny*SparkZ2 +Nx*SparkY2 + SparkX2]<<" "<< cj[Nx*Ny*SparkZ2 +Nx*SparkY2 + SparkX2]<< " "<< Nr[Nx*Ny*SparkZ2 +Nx*SparkY2 + SparkX2]<< " "<< Nrc[Nx*Ny*SparkZ2 +Nx*SparkY2 + SparkX2]<< " " << Nri[Nx*Ny*SparkZ2 +Nx*SparkY2 + SparkX2] << " " << Nrr[Nx*Ny*SparkZ2 +Nx*SparkY2 + SparkX2] << " " << Nl[Nx*Ny*SparkZ2 +Nx*SparkY2 + SparkX2]<< " " << Nlc[Nx*Ny*SparkZ2 +Nx*SparkY2 + SparkX2]<< " " << Nlc2[Nx*Ny*SparkZ2 +Nx*SparkY2 + SparkX2]<< " " << Nli[Nx*Ny*SparkZ2 +Nx*SparkY2 + SparkX2] << " " << Nli2[Nx*Ny*SparkZ2 +Nx*SparkY2 + SparkX2]<< " " << Nlb1[Nx*Ny*SparkZ2 +Nx*SparkY2 + SparkX2]<< " " << Nlb2[Nx*Ny*SparkZ2 +Nx*SparkY2 + SparkX2]<<endl;
 
 #else
 
 singlesparkfile << t << " "<< cd[Nx*Ny*SparkZ +Nx*SparkY + SparkX]<< " " << Nr[Nx*Ny*SparkZ +Nx*SparkY + SparkX] <<endl;

 #endif
 
#endif


#ifdef BCLUSTERTIME
 if(t>=BCLUSTERTIME && t<=ECLUSTERTIME && step % clusterstep == 0 ){

 ///////////FOR CLUSTER///////////////////////////

    for (int k=0; k<Nz; k++){
     for (int i=0; i<Ny; i++){
      for (int j=0; j<Nx; j++){
       int idx=Nx*Ny*k +Nx*i + j;
       //if (dfun[idx] == 1) {
        if(idx==Nx*Ny*(Nz-1)+Nx*(Ny-1)+(Nx-1)){clusterfile << cm[idx]<<endl;} else {clusterfile << cm[idx]<< " ";};
       //};
      };
     };
    };
  ////////////////////////////////////////////////

 };
#endif

#ifdef BMOVIETIME
 if(t>=bmovietime1 && t<=emovietime1 && step % moviestep == 0 ){

  
  ////////// FOR MOVIES //////////////////////////
 #ifdef SR_MOVIE
   write_frame(t, cs);
 #else
   write_frame(t, cm);
 #endif
   dump_data_to_files();
  ////////////////////////////////////////////////

 };
#endif 

};  // end (if step % outstep == 0)

#endif
//////// end output /////////////////////////////




/////////////// solving equations ///////////////////

////////// diffusion steps for outside loop
for (int kd=0; kd<Ndiff; kd++){
//printf("time %f v %f\n", t, x[0]); fflush(stdout);

//flux (and channel) steps
#ifdef DO_FLUX
for (int kf=0; kf<Nflux_Ndiff; kf++){

////// update channels


#ifdef USE_LCC_RAND
//#if defined(TTdisrupt) || defined(plcc)
// Lcc7stateUpdate_TTdisrupt <<< numBlocks, threadsPerBlock >>> (cd_d, cm_d, nr_d, Nl_d, Nlc_d, Nlc2_d, Nli_d, Nli2_d, Nlb1_d, Nlb2_d, lfire_d, dfun_d, seed_d1,seed_d2,seed_d3,seed_d4, dtflux, v, NxNyNz, t);
//#else
 Lcc7stateUpdate <<< numBlocks, threadsPerBlock >>> (cd_d, Nl_d, Nlc_d, Nlc2_d, Nli_d, Nli2_d, Nlb1_d, Nlb2_d, lfire_d, dfun_d, seed_d1,seed_d2,seed_d3,seed_d4, dtflux, v, NxNyNz, t);
//#endif
#endif


#ifdef USE_RYR_ODE
 RyrUpdate_ODE<<< numBlocks, threadsPerBlock >>>(cd_d, Nr_d, Nrc_d, Nri_d, Nrr_d , Pro_d, Pri_d, Prc_d, dfun_d, dtflux, NxNyNz);
 cudaThreadSynchronize();
#endif


#ifdef USE_RYR_RAND
 RyrUpdate <<< numBlocks, threadsPerBlock >>> (cd_d, Nr_d, Nrc_d, Nri_d, Nrr_d ,rfire_d, dfun_d,seed_d1,seed_d2,seed_d3,seed_d4, dtflux, NxNyNz);
#endif

cudaThreadSynchronize();//synch threads after stochastic channel updates

/////// update flux

#ifdef DO_UCLA_AP
#ifndef CLAMP_NAI
 Nai3value=(x[19]*x[19]*x[19]);
#endif
#endif

#ifdef USE_RYR_ODE
 if (usetemp == 1){
 FluxUpdate_RYR_ODE <<< numBlocks, threadsPerBlock >>> (cmtemp_d, cstemp_d, cj_d, cd_d, dfun_d, Pro_d, Nl_d, nr_d, Ilcc_d, Iryr_d, Jncx_d, Nai3value, v, dtflux, NxNyNz, t);
}
else {
 FluxUpdate_RYR_ODE <<< numBlocks, threadsPerBlock >>> (cm_d, cs_d, cj_d, cd_d, dfun_d, Pro_d, Nl_d, nr_d, Ilcc_d, Iryr_d, Jncx_d, Nai3value, v, dtflux, NxNyNz, t);
};


#else

if (usetemp == 1){
 FluxUpdate <<< numBlocks, threadsPerBlock >>> (cmtemp_d, cstemp_d, cj_d, cd_d, dfun_d, Nr_d, Nl_d, nr_d, Ilcc_d, Iryr_d, Jncx_d, Nai3value, v, dtflux, NxNyNz, t);
}
else {
 FluxUpdate <<< numBlocks, threadsPerBlock >>> (cm_d, cs_d, cj_d, cd_d, dfun_d, Nr_d, Nl_d, nr_d, Ilcc_d, Iryr_d, Jncx_d, Nai3value, v, dtflux, NxNyNz, t);
};

#endif
cudaThreadSynchronize();


};
#endif
//ends fluxsteps


//start diffusion steps

#ifdef DO_DIFFUSION

#ifdef BLOCK_SIZE_X

 compute_pde2<<<dimGrid, dimBlock>>>(cm_d, cmtemp_d, DxMdt_dx2, Nx, Ny, Nz, NxNy);
 cudaThreadSynchronize();
 // ~ for (int kz=0; kz<Nz; kz++){
   // ~ 
  // ~ compute_pde<<<dimGrid, dimBlock>>>(cm_d, cmtemp_d, DxMdt_dx2, Nx, Ny, Nz, kz, NxNy, (NxNy*kz) );
  // ~ cudaThreadSynchronize();
 // ~ 
 // ~ };
 // ~ 

#else

 if (usetemp == 1){
  
  #ifdef USE_TEX
   tex_myo_diffusion_finitediff_GPU <<< numBlocks, threadsPerBlock >>> (cmtemp_d, cm_d, DxMdt_dx2, DyMdt_dx2, DzMdt_dx2, Nx, Ny, Nz, NxNy, NxNyNz, usetemp);
  #else
   myo_diffusion_finitediff_GPU <<< numBlocks, threadsPerBlock >>> (cmtemp_d, cm_d, DxMdt_dx2, DyMdt_dx2, DzMdt_dx2, Nx, Ny, Nz, NxNy, NxNyNz);
  #endif
  
  #ifndef FIX_SR
  #ifdef USE_TEX
   tex_sr_diffusion_finitediff_GPU <<< numBlocks, threadsPerBlock >>> (cstemp_d, cs_d, DxSdt_dx2, DySdt_dx2, DzSdt_dx2, Nx, Ny, Nz, NxNy, NxNyNz, usetemp);
  #else
   sr_diffusion_finitediff_GPU <<< numBlocks, threadsPerBlock >>> (cstemp_d, cs_d, DxSdt_dx2, DySdt_dx2, DzSdt_dx2, Nx, Ny, Nz, NxNy, NxNyNz);
  #endif
  #endif
  usetemp=0;
 }
 else{
  
  #ifdef USE_TEX
   tex_myo_diffusion_finitediff_GPU <<< numBlocks, threadsPerBlock >>> (cm_d, cmtemp_d, DxMdt_dx2, DyMdt_dx2, DzMdt_dx2, Nx, Ny, Nz, NxNy, NxNyNz, usetemp);
  #else
   myo_diffusion_finitediff_GPU <<< numBlocks, threadsPerBlock >>> (cm_d, cmtemp_d, DxMdt_dx2, DyMdt_dx2, DzMdt_dx2, Nx, Ny, Nz, NxNy, NxNyNz);
  #endif
  #ifndef FIX_SR
  #ifdef USE_TEX
   tex_sr_diffusion_finitediff_GPU <<< numBlocks, threadsPerBlock >>> (cs_d, cstemp_d, DxSdt_dx2, DySdt_dx2, DzSdt_dx2, Nx, Ny, Nz, NxNy, NxNyNz, usetemp);
  #else
   sr_diffusion_finitediff_GPU <<< numBlocks, threadsPerBlock >>> (cs_d, cstemp_d, DxSdt_dx2, DySdt_dx2, DzSdt_dx2, Nx, Ny, Nz, NxNy, NxNyNz);
  #endif
  #endif
  usetemp=1;

 };

#endif


#ifdef UNFIX_SR_TIME
if (t>= UNFIX_SR_TIME){sr_diffusion_finitediff_GPU <<< numBlocks, threadsPerBlock >>> (cs_d, cstemp_d, DxSdt_dx2, DySdt_dx2, DzSdt_dx2, Nx, Ny, Nz, NxNy, NxNyNz);};
#endif

cudaThreadSynchronize();//synch after diffusion steps

// ~ 
// ~ //update cm_d and cs_d below
// ~ 
// ~ stateUpdate <<< numBlocks, threadsPerBlock >>> (cm_d,cmtemp_d,NxNyNz);
// ~ 
// ~ #ifndef FIX_SR
  // ~ stateUpdate <<< numBlocks, threadsPerBlock >>> (cs_d,cstemp_d,NxNyNz);
// ~ #endif
// ~ 
// ~ #ifdef UNFIX_SR_TIME
// ~ if (t>= UNFIX_SR_TIME){stateUpdate <<< numBlocks, threadsPerBlock >>> (cs_d,cstemp_d,NxNyNz);};
// ~ #endif
// ~ 
// ~ cudaThreadSynchronize();


#endif


//ends diffusion steps

};//ends diffusion outside loop

gpuErrchk(cudaDeviceSynchronize());

/// do AP update steps

///start UCLA AP steps

#define DO_UCLA_AP
#ifdef DO_UCLA_AP

 double cmavg;
 double csavg; 
 if (usetemp == 1){
 //cudaMemcpy(cm, cmtemp_d, size, cudaMemcpyDeviceToHost);
 cmavg = computeReduction(NxNyNz, cmtemp_d, SUMdumb_d);
 csavg = computeReduction(NxNyNz, cstemp_d, SUMdumb_d);
 }
 else{
 //cudaMemcpy(cm, cm_d, size, cudaMemcpyDeviceToHost);
 cmavg = computeReduction(NxNyNz, cm_d, SUMdumb_d);
  csavg = computeReduction(NxNyNz, cs_d, SUMdumb_d);
 };

 double cdavg = computeReduction(NxNyNz, cd_d, SUMdumb_d); 
 double cjavg = computeReduction(NxNyNz, cj_d, SUMdumb_d); 
 
 //double Nlavg = computeReduction(size, Nl_d);
 double Jncxavg = computeReduction(NxNyNz, Jncx_d, SUMdumb_d);
 double Ilccavg = computeReduction(NxNyNz, Ilcc_d, SUMdumb_d);

 /*cudaMemcpy(cd, cd_d, size, cudaMemcpyDeviceToHost);
 cudaMemcpy(Nl, Nl_d, sizeint, cudaMemcpyDeviceToHost);
 cudaMemcpy(Jncx, Jncx_d, size, cudaMemcpyDeviceToHost);
 cudaMemcpy(Ilcc, Ilcc_d, size, cudaMemcpyDeviceToHost);

 double cmavg=0.0;
 double cdavg=0.0;
 double Nlavg=0.0;
 double Jncxavg=0.0;
 double Ilccavg=0.0;

 for (int k=0; k<Nz; k++){
  for (int i=0; i<Ny; i++){
   for (int j=0; j<Nx; j++){
     int idx=Nx*Ny*k +Nx*i + j;
     cmavg+=cm[idx];
     Nlavg+=Nl[idx];
     cdavg+=cd[idx];
     Jncxavg+=Jncx[idx];
     Ilccavg+=Ilcc[idx];
    };
   };
 };
 */

 cmavg=cmavg/(NxNyNz);
 Jncxavg=Jncxavg/(NxNyNz);
 Ilccavg=Ilccavg/crucount;
 cdavg=cdavg/crucount;
 csavg = csavg/NxNyNz;
 cjavg = cjavg/crucount;
 //Nlavg=Nlavg/crucount;

//debug
 // Ilccavg = 0;
// Jncxavg = 0;
 //cmavg = 0;

 #ifdef USE_LCC_CELL

   x[28]=2.0*wca*(-Ilccavg);//to use my computed lcc flux 
 #else
 /* const FLOAT phi=v*F_RT;
  IlccTotal=(Nlavg/Nlcc)*182*4.0*0.00054*v*F*F_RT*(-Bca*1.8+exp(2*phi)*cdavg/1000.0)/(exp(2*phi)-1);
   //IlccTotal=(x[30])*182*4.0*0.00054*v*F*F_RT*(-Bca*1.8+exp(2*phi)*cdavg/1000.0)/(exp(2*phi)-1);
  x[28]=2.0*wca*IlccTotal;//to use ucla lcc flux with my Nl/Nlcc for Po
 */#endif



 x[27]=wca*Jncxavg;
 x[29]=cmavg;


 tmod= (t-int(t/uclaPCL)*uclaPCL);
 if (0<= tmod &&  tmod <= DT_STIM ) {
     ucla_one_step(dt, x, I_STIM, NULL); 
     
 }
 else {
   ucla_one_step(dt, x, 0.0, NULL); 
 };
 //x[0] = -86.0f;
 //printf("time %f v %f\n", t, x[0]); fflush(stdout);

 fprintf(outFile,"%f %f %f %f %f %f %f %f %f\n", t, x[0], x[29], Jncxavg, Ilccavg, cdavg, csavg, cjavg, x[19]);

#ifndef AP_CLAMP
 v=x[0];
#else
 x[0]=v;
#endif

#endif
///////// end UCLA AP 


///////////////////////////update step and time (after dynamic AP, but before fixed AP)
step++;
t=dt*step;
/////////////////////////////////////////////////

/// fixed AP steps

#ifdef DO_FIXED_AP
#ifdef NO_AP_RAMP
tmod=t-int(t/PCL)*PCL;
tempnum2=PCL-APD;
if (tmod<tempnum2) {v=vrest;}
else {v=vap;}
#else
#ifdef STOP_FIXED_AP
if(t<STOP_FIXED_AP){
tmod=t-int(t/PCL)*PCL;
tempnum1=.75*APD;
tempnum2=PCL-APD;
tempnum3=tempnum2+tempnum1;
if (tmod<tempnum2) {v=vrest;}
else if (tmod >= tempnum2 && tmod <= tempnum3) {v=vap;}
else {v=-4.0*(vap-vrest)*tmod/APD+4*PCL*(vap-vrest)/APD+vrest;};
}
else{
v=vrest;
};
#else
tmod=t-int(t/PCL)*PCL;
tempnum1=.75*APD;
tempnum2=PCL-APD;
tempnum3=tempnum2+tempnum1;
if (tmod<tempnum2) {v=vrest;}
else if (tmod >= tempnum2 && tmod <= tempnum3) {v=vap;}
else {v=-4.0*(vap-vrest)*tmod/APD+4*PCL*(vap-vrest)/APD+vrest;};
#endif
#endif
#endif

#ifdef AP_CLAMP
 #ifdef STOP_FIXED_AP
  if(t<STOP_FIXED_AP){
   v=v_apclamp[int((t-int(t/PCL)*PCL)/AP_OUTTIME)];
  }
  else{
   v=vrest;
  };
 #else
  aptmod=int((t-int(t/PCL)*PCL)/AP_OUTTIME);
  if (aptmod<length_v_apclamp) {
   v=v_apclamp[aptmod];
  }
  else{
   v=vaprest;
  };
 #endif
 #ifdef DO_UCLA_AP
  x[0]=v;
 #endif
#endif


cudaError_t err = cudaGetLastError();
 if( err != cudaSuccess ) {
    printf("Error: %s\n", cudaGetErrorString(err) );
    assert(err== cudaSuccess);
 };



};

/*
cutStopTimer( timer);
cout<< "end"<<" "<< cutGetTimerValue( timer)/1000 << " seconds" <<" "<< avgfile << " "<< "DevNum" << " " << DevNum <<endl;
cutDeleteTimer( timer);
*/

free(cm);free(cmtemp);free(cs);free(cstemp);free(cd);free(cdtemp);free(cj);free(dfun);free(nr);free(Nr);free(Nri);free(Nrc);free(Nrr);
free(nl);free(Nl);free(Nli);free(Nlc);free(rfire);free(lfire);free(Jncx);free(Jup);free(FluxM);free(FluxS);free(Jleak);free(FluxD);free(FluxJ);free(Ilcc);
free(seed_h1);free(seed_h2);free(seed_h3);free(seed_h4);free(Nlb1);free(Nlb2);free(Nli2);free(Nlc2);

cudaFree(cm_d);cudaFree(cmtemp_d);cudaFree(cs_d);cudaFree(cstemp_d);cudaFree(cd_d);cudaFree(cdtemp_d);cudaFree(cj_d);cudaFree(dfun_d);cudaFree(Nr_d);
cudaFree(Nl_d);cudaFree(Nli_d);cudaFree(Nlc_d);cudaFree(lfire_d);cudaFree(Nrr_d);cudaFree(Nri_d);cudaFree(Nrc_d);cudaFree(rfire_d);
cudaFree(seed_d1);cudaFree(seed_d2);cudaFree(seed_d3);cudaFree(seed_d4);cudaFree(Nlb1_d);cudaFree(Nlb2_d);cudaFree(Nli2_d);cudaFree(Nlc2_d);

fclose(outFile);

}





////////// FOR MOVIES //////////////////////////
#ifdef BMOVIETIME
static void write_frame(FLOAT t, FLOAT *x)
{  
  int i, j, k;
  FLOAT v, v_min = FLT_MAX, v_max = -FLT_MAX;
  FLOAT maxchar = UCHAR_MAX + 0.0;
  
  // Determine the range of the data
  for (i = 0; i < Ny; i+= DY_SAMPLE) {
    for (j = 0; j < Nx; j+= DX_SAMPLE) {
      v = x[Nx*Ny*MovieZ+i*Nx+j];
      v_min = v < v_min ? v : v_min;
      v_max = v > v_max ? v : v_max;       
    }   
  }
  
  // Handle singluarities.
  if (((v_max)-(v_min)) < MIN_VAR_DIFF)
    v_max = v_min + MIN_VAR_DIFF;
  
  // Save the data range
  vm_bounds[0][0] = t;
  vm_bounds[0][1] = v_min;
  vm_bounds[0][2] = v_max;
  
  // Put the values of the plot variables in the buffer.
#ifdef _OPENMP
#pragma omp parallel for shared(x, vm_frame_buffer) private(i,j,k)
#endif
  for (i = 0; i < Ny; i+= DY_SAMPLE) {
    for (j = 0; j < Nx; j+= DX_SAMPLE) {
      k = j/DX_SAMPLE + i/DY_SAMPLE*Nx/DX_SAMPLE;
      vm_frame_buffer[0][k] = (unsigned char) ( (x[Nx*Ny*MovieZ+i*Nx+j]-vm_bounds[0][1])/(vm_bounds[0][2]-vm_bounds[0][1])*maxchar );
    }
  }
  
}

/**
 * \brief Dumps all buffers to the output files. And resets the counters.
 */
static void dump_data_to_files(void)
{
  // Dump the frame data.
    // Bounds
    fprintf(fptr_bounds, "%E\t%E\t%E\t", vm_bounds[0][0], vm_bounds[0][1], 
            vm_bounds[0][2]);
    fprintf(fptr_bounds, "\n");
    // Binary data
    fwrite(vm_frame_buffer[0], sizeof(unsigned char), REC_N_CELLS, fptr_video);

}
#endif
///////////////////////////////////////





 


